#include "icp.h"

#include <iostream>

#include <Eigen/Dense>

using FloatType = double;
using IndexType = int64_t;
using EigenMatrix = Eigen::MatrixXd;
using EigenVector = Eigen::VectorXd;

#include "common/io.h"
#include "common/kdtree.h"
#include "common/debug.h"
#include "svd.h"

// Custom struct for KD tree
struct Point : public Vec3 {
    Point() {}
    Point(const Vec3 &v, int index = -1) : Vec3(v), i(index) {}
    int i;
};

// Rodrigues rotation matrix
EigenMatrix rodrigues(const EigenVector &w, const double theta) {
    const double c = std::cos(theta);
    const double s = std::sin(theta);

    const double wx = w(0);
    const double wy = w(1);
    const double wz = w(2);
    EigenMatrix K(3, 3);
    K << 0.0, -wz, wy, wz, 0.0, -wx, -wy, wx, 0.0;

    EigenMatrix I = EigenMatrix::Identity(3, 3);
    return I + K * s + (K * K) * (1.0 - c);
}

//! single point to point ICP step
void point2pointICP_step(const std::vector<Vec3> &target, const std::vector<Vec3> &source, EigenMatrix *rotMat,
                         EigenVector *trans) {
    // Construct Kd-tree to find closest points
    KDTree<Vec3> tree;
    tree.construct(target);

    NOT_IMPL_ERROR();

    // Match closest point pairs
    // Hint:
    // Construct here that two matrices X and P,
    // whose elements are positions of source vertices (X),
    // and those for closest points of the sources (P)

    // Subtract xMean, pMean from X, P
    // Hint:
    // Row-wise or column-wise mean (also sum, max, min etc.) of Eigen matrices
    // can be easily calculate using "rowwise()" or "colwise()".

    // Solve with SVD to obtain R
    // Hint:
    // An SVD method "eigenSVD" is already given in "svd.h".
    // See its arguments carefully when using it.

    // Revise sign of determinant

    // Output
    // Hint:
    // Be careful that arguments for output matrices and vectors
    // are represented as "pointers". Therefore, you should insert
    // values for them with "asterisk" like "*lhs = rhs".

}

//! single point to plane ICP step
void point2planeICP_step(const std::vector<Vec3> &target, const std::vector<Vec3> &targetNorm,
                         const std::vector<Vec3> &source, Eigen::MatrixXd *rotMat, Eigen::VectorXd *trans) {
    // Construct Kd-tree to find closest points
    std::vector<Point> points;
    for (int i = 0; i < (int)target.size(); i++) {
        points.emplace_back(target[i], i);
    }

    KDTree<Point> tree;
    tree.construct(points);

    NOT_IMPL_ERROR();

    // Construct linear system
    // Hint:
    // prepare matrix A and vector b
    // A is 6x6 matrix, and b is 6-D vector

    // Solve
    // Hint:
    // Use "Eigen::PartialPivLU" to solve the system

    // Substitute to instances a, t
    // Hint:
    // Eigen matrices and vectors can be initialized
    // with its elements by "<<" operator.

    // Reproduce rotation matrix
    // Hint:
    // Remember that vector "a" is the multiple of
    // the direction of rotation axis and
    // the degree of rotation angle.

}

void rigidICP(const std::vector<Vec3> &target, const std::vector<Vec3> &targetNorm, std::vector<Vec3> &source,
              std::vector<Vec3> &sourceNorm, ICPMetric metric, int maxIters, double tolerance, bool verbose) {
    // Point-to-point ICP
    for (int it = 0; it < maxIters; it++) {
        Eigen::MatrixXd R(3, 3);
        Eigen::VectorXd t(3);
        switch (metric) {
        case ICPMetric::Point2Point:
            point2pointICP_step(target, source, &R, &t);
            break;

        case ICPMetric::Point2Plane:
            point2planeICP_step(target, targetNorm, source, &R, &t);
            break;

        default:
            throw std::runtime_error("Unknown ICP metric type!");
        }

        // Apply rigid transformation
        for (int i = 0; i < (int)source.size(); i++) {
            Eigen::VectorXd v(3);
            v << source[i].x, source[i].y, source[i].z;
            Eigen::VectorXd u = R * v + t;
            source[i] = Vec3(u(0), u(1), u(2));

            Eigen::VectorXd n(3);
            n << sourceNorm[i].x, sourceNorm[i].y, sourceNorm[i].z;
            Eigen::VectorXd m = R * n;
            sourceNorm[i] = Vec3(m(0), m(1), m(2));
        }

        // Check current error
        const double error = t.norm() + (R - Eigen::MatrixXd::Identity(3, 3)).norm();

        // Report
        if (verbose) {
            printf("*** %d iteration ***\n", it + 1);
            printf("R = \n");
            std::cout << R << std::endl;
            printf("t = \n");
            std::cout << t << std::endl;
            printf("error = %f\n", error);
            printf("\n");
        }

        char outfile[256];
        sprintf(outfile, "intermediate_%03d.off", it);
        write_off(std::string(outfile), source, sourceNorm);

        if (error < tolerance) {
            break;
        }
    }
}
